from enum import Enum
import inquirer


# author type
class AuthorType(Enum):
    STUDENT = "student"
    FACULTY = "faculty"
    # GROUP = "Group"


# author
# TODO: Add a group type which holds a list of members
class Author:
    # --- CLASS MEMBERS (INSTANCE VARIABLES) ---
    # type of author
    type: AuthorType
    # author name
    name: str

    # --- CLASS VARIABLES
    NAME = "Name"
    TYPE = "Type"


    # constructor
    def __init__(self, type: AuthorType, name: str) -> None:
        self.type = type
        self.name = name

    def to_dict(self):
        return {
            Author.NAME: self.name,
            Author.TYPE: AuthorType(self.type).value
        }

    # create an author object
    @staticmethod
    def create_author():
        author_query = inquirer.List(
            'author_type',
            'Select the type of author you want',
            [atype.value for atype in AuthorType]
        )
        author_name = inquirer.Text(
            'author_name',
            'Enter the name of the {author_type}'
        )
        answer = inquirer.prompt([author_query, author_name])
        try:
            return Author(
                AuthorType(
                    answer['author_type']
                ),
                answer['author_name']
            )
        except TypeError as te:
            print(te)
            raise te




from typing import List

MAKEFILE_PHONY = ".PHONY: all clean"
PHONY_BUILD_ALL = "all:"
PHONY_CLEAN_ALL = "clean:"

COMPILE_COMMAND: str = "\tiverilog -o $@ $^"
CLEAN_COMMAND: str = "\t-rm $({}) $({})"
EXECUTABLE_COMMAND: str = "\ttimeout {}s vvp $<"
EXECUTABLE_WITHOUT_TIMEOUT_COMMAND = "vvp $<"
REPORT_COMMAND: str = "\t@printf \"\\e[0;94m{}\\e[0m\\n\""
VIEWER_COMMAND: str = "\tgtkwave -6 $<"
MAKEFILE_MODULE_COMMAND: str = "\tmake {}"

CONST_INPUT_COMMAND: str = "{}:\t$({})"
INPUT_OUTPUT_COMMAND: str = "$({}):\t$({})"

def gen_executable(timeout_in_seconds: int = 10):
    if timeout_in_seconds <= 0:
        return EXECUTABLE_WITHOUT_TIMEOUT_COMMAND
    else:
        return EXECUTABLE_COMMAND.format(str(timeout_in_seconds))

def gen_report(outstring: str = ""):
    return REPORT_COMMAND.format(outstring)

def gen_inputvar_const(inputvar: str = "", constvar: str = ""):
    return CONST_INPUT_COMMAND.format(constvar, inputvar)

def gen_input_output(inputvar: str = "", outputvar: str = ""):
    return INPUT_OUTPUT_COMMAND.format(outputvar, inputvar)

def gen_report_cmd(mod_name: str = "", message: str = ""):
    return REPORT_COMMAND.format(f"{mod_name}: \\e[0;92m{message}")

def gen_makefile_module_cmd(module_name: str):
    return MAKEFILE_MODULE_COMMAND.format(module_name)

def append_lines(*commands: List[str]) -> str:
    final_output = ""
    for command in commands:
        final_output += command + "\n"
    return final_output



from glob import glob
import os
import pathlib
from string import Template
from typing import Any


def create_template(
    var: str,
    module: Any,
    subproject: str,
    project: str
) -> str:
    return Template(var).safe_substitute(
        name=module.name,
        exec=module.exec_file,
        wave=module.wave_file,
        buildcommand=module.build_command,
        cleancommand=module.clean_command,
        subproject=subproject,
        project=project
    )

def generate_quickstart_template(module: Any, subproject: str, project: str) -> list[str]:
    templates = glob(pathlib.Path(pathlib.Path(os.getcwd()).parts[:-1]),"*.v.lexicon_qs")
    generated_files = []
    for file in templates:
        new_filename = create_template(file, module, subproject, project)[:-11]
        generated_files.append(new_filename)
        with open(file, "r") as infile:
            with open(new_filename, "w") as outfile:
                outfile.write(create_template(infile.read(), module, subproject, project))
    return generated_files
    

# generate_template(newmod, "A1Q01", "Assignment_01")


#!/usr/bin/python3
# This holds the information regarding the students for the project needed.

from copy import deepcopy
from enum import Enum
import colorama
import inquirer
import os
from typing import Dict, List, Any, Type
import yaml


from lexicon.commonutils import CLI_VARS_PRETTYPRINT, change_text, check_is_modified

from lexicon.author import Author, AuthorType
from lexicon.commonutils import check_is_modified
from lexicon.subproject import Subproject

class ProjectType(Enum):
    ASSIGNMENT = "assignment"
    EXAMINATION = "examination"
    PROJECT = "project"
    PERSONAL = "personal"
    FACULTY_COPY = "faculty copy"
    ANSWER_KEY = "answer key"
    OTHER = "other"

class Project:
    # -- CLASS MEMBERS (INSTANCE VARIABLES) ---
    # project name
    name: str
    # type of project
    type: ProjectType
    # author of the project
    author: Author
    # subprojects
    subprojects: List[str]

    # PROJECT FILE
    PROJECT_FILE = "lexicon.yaml"
    # --- CLASS CONSTANTS ---
    # name of the project
    NAME = "Name"
    # this holds the type of project, check ProjectType class
    TYPE_OF_PROJECT = "Type"
    # holds the list of subprojects, organized in directories
    # in a project directory
    SUBPROJECTS = "Subprojects"
    # the author of the project, can be faculty or student
    AUTHOR = "Author"

    def __init__(self, args: Dict[str, Any]) -> None:
        self.name = args[Project.NAME].strip()
        self.type = ProjectType(args[Project.TYPE_OF_PROJECT].strip())
        author = args[Project.AUTHOR]
        self.author = Author(AuthorType(author[Author.TYPE].lower()), author[Author.NAME])
        self.subprojects = args[Project.SUBPROJECTS]
        

    def to_dict(self):
        return {
            Project.NAME: self.name,
            Project.TYPE_OF_PROJECT: ProjectType(self.type).value,
            Project.AUTHOR: self.author.to_dict(),
            Project.SUBPROJECTS: self.subprojects
        }

    def to_yaml(self):
        return yaml.safe_dump(self.to_dict())

    def save_config(self):
        project_path = os.path.join(os.getcwd(), Project.PROJECT_FILE)
        with open(project_path, "w") as new_config_file:
            yaml.safe_dump(self.to_dict(), new_config_file)


    # --- CLI METHODS ---
    # generate projects
    def cli_generate():
        do_create = inquirer.confirm("Do you want to create the project file right now?")
        if do_create:
            # questions to prompt the user
            project_queries = [
                inquirer.Text(
                    'name',
                    'Enter the name of the project'
                ),
                inquirer.List(
                    'type',
                    'Enter the type of project',
                    list(ProjectType)
                )
            ]
            # ask questions and get answers
            answer = inquirer.prompt(project_queries)
            # create an author from CLI prompt
            author: Author = Author.create_author()
            # subprojects list, which is empty presently
            subprojects = []
            spcreateprompt = inquirer.Text(
                'name',
                'Enter the name of the new subproject'
            )
            # add any generated projects here, for each generated project, create
            # an empty directory also
            while True:
                if subprojects == []:
                    if inquirer.confirm("No subprojects created. Create one?"):
                        spname = inquirer.prompt([spcreateprompt])['name']
                        if inquirer.confirm(
                            f"Do you want to continue creating '{spname}'?"
                        ):
                            sp_pathname = os.path.join(os.getcwd(), spname)
                            if not os.path.exists(sp_pathname):
                                os.makedirs(spname)
                            subprojects.append(spname)
                            Subproject.generate_subproject(
                                spname,
                                answer['name'],
                                author.name
                            )
                    else:
                        break
                else:
                    if inquirer.confirm("Do you want to create more subprojects?"):
                        spname = inquirer.prompt([spcreateprompt])['name']
                        if inquirer.confirm(
                            f"Do you want to continue creating {spname}?"
                        ):
                            os.makedirs(os.path.join(os.getcwd(), spname))
                            subprojects.append(spname)
                            Subproject.generate_subproject(
                                spname, 
                                answer['name'],
                                author.name
                            )
                    else:
                        break
                        
            # return the generated project
            try:
                project_config = Project({Project.AUTHOR: author.to_dict(), Project.NAME: spname, Project.TYPE_OF_PROJECT: answer['type'].lower(), Project.SUBPROJECTS: subprojects})
                with open(os.path.join(os.getcwd(), Project.PROJECT_FILE), 'w') as project_out:
                    project_out.write(project_config.to_yaml())
                    print("Project configuration written")
            except TypeError as te:
                print(te)
                return None

    # --  VARIABLES FOR STATIC METHOD --
    CHOICE = "choice"

    # manage projects
    @staticmethod
    def cli_manage():
        project_path = os.path.join(os.getcwd(), Project.PROJECT_FILE)
        if not os.path.exists(project_path):
            Project.cli_generate()
        else:
            project_config = yaml.safe_load(open(project_path, "r"))
            project_config: 'Project' = Project(project_config)
            print(project_config.to_dict())
            current_config = deepcopy(project_config)
            is_modified = False
            # CLI Loop
            while True:
                root_prompt = inquirer.List(
                    Project.CHOICE,
                    message = f"Manage {colorama.Fore.YELLOW}Project {colorama.Fore.BLUE}{project_config.name} {check_is_modified(is_modified, f'{colorama.Fore.GREEN}<modified>')}{colorama.Fore.RESET}",
                    choices = ["Manage Subprojects", "Modify Properties", "Reset", "View Configuration", "Exit"]
                )
                try:
                    choice = inquirer.prompt([root_prompt])[Project.CHOICE]
                except TypeError:
                    continue
                if choice == None:
                    continue
                elif choice == "Exit":
                    if is_modified:
                        modify_write = inquirer.confirm("Project has been modified. Do you want to save changes?")
                        print(f"UMODIFIED:{current_config.to_dict()}")
                        print(f"MODIFIED: {project_config.to_dict()}")
                        if modify_write:
                            project_config.save_config()
                            print(f"{colorama.Fore.GREEN} Configuration file saved in filepath {colorama.Fore.YELLOW}{os.path.join(os.getcwd(), Project.PROJECT_FILE)}{colorama.Fore.RESET} .")
                    break
                elif choice == "Reset":
                    project_config = deepcopy(current_config)
                    is_modified = False
                elif choice == "Manage Subprojects":
                    # SUBPROJECT LOOP
                    while True:
                        subproject_choices = list(project_config.subprojects) + [
                            "Create a subproject",
                            "Exit"
                        ]
                        subproject_prompt = inquirer.List(
                            'subproject_choice',
                            'Select Subproject to manage',
                            choices = subproject_choices
                        )
                        subproject_choice = inquirer.prompt([subproject_prompt])['subproject_choice']
                        if subproject_choice == "Exit":
                            break
                        elif subproject_choice == "Create a subproject":
                            # USE Subproject.generate_subproject to create a subproject here
                            # and add it to project_config.subprojects
                            spcreateprompt = inquirer.Text(
                                'subproject_name',
                                'Enter the name for the subproject'
                            )
                            spname = inquirer.prompt([spcreateprompt])
                            if spname != None:
                                spname = spname['subproject_name']
                                if spname.strip() != "":
                                    os.makedirs(os.path.join(os.getcwd(), spname))
                                    project_config.subprojects.append(spname)
                                    Subproject.generate_subproject(
                                        spname,
                                        project_config.name,
                                        project_config.author.name
                                    )
                                    is_modified = True
                                else:
                                    print(f"{colorama.Fore.RED} NO NAME GIVEN. IGNORING.{colorama.Fore.RESET}")
                                    break
                            else:
                                break
                        else:
                            subproject_name = subproject_choice
                            print(f"{colorama.Fore.BLUE}Switching to subproject {colorama.Fore.YELLOW}{subproject_choice}{colorama.Fore.RESET}")
                            subproject_path = os.path.join(os.path.dirname(project_path), subproject_name)
                            print(subproject_path)
                            if os.path.exists(subproject_path):
                                sub_is_modified = Subproject.transfer_control(subproject_name)
                                if sub_is_modified:
                                    is_modified = True
                            else:
                                print(f"{colorama.Fore.RED}PATH DOES NOT EXIST{colorama.Fore.RESET}")
                                break
                elif choice == "Modify Properties":
                    # Choice for Project Properties
                    project_dict = project_config.to_dict()
                    keydictionary = {}
                    for var in [
                        Project.NAME,
                        Project.TYPE_OF_PROJECT,
                        # Project.AUTHOR
                    ]:
                        keydictionary[CLI_VARS_PRETTYPRINT.format(var, project_dict[var])] = var
                    keydictionary[CLI_VARS_PRETTYPRINT.format(Project.AUTHOR, project_dict[Project.AUTHOR][Author.NAME])] = Project.AUTHOR
                    choice_list = inquirer.List(
                        'project_key',
                        message='Select the property to modify',
                        choices= list(keydictionary.keys())
                    )
                    try:
                        choice_property = keydictionary[inquirer.prompt([choice_list])['project_key']]
                    except TypeError:
                        continue
                    if choice_property == None:
                        continue
                    elif choice_property == Project.NAME:
                        result, answer = change_text(choice_property, project_dict[choice_property])
                        if result:
                            project_config.name = answer
                        else:
                            break
                    elif choice_property == Project.TYPE_OF_PROJECT:
                        # let user select the type of project the user needs
                        type_choices_prompt = inquirer.List(
                            'project_type',
                            'Select the property type you want to switch to',
                            choices = [ptype.value for ptype in ProjectType],
                            default = project_config.type
                        )
                        try:
                            ptype_choice = inquirer.prompt([type_choices_prompt])['project_type']
                            # project type changed here
                            project_config.type = ProjectType(ptype_choice)
                        except TypeError:
                            continue
                    elif choice_property == Project.AUTHOR:
                        author = project_config.author
                        author_dict = author.to_dict()
                        keydictionary = {}
                        keydictionary[CLI_VARS_PRETTYPRINT.format(Author.NAME, author.name)] = Author.NAME
                        keydictionary[CLI_VARS_PRETTYPRINT.format(Author.TYPE, author.type.value)] = Author.TYPE
                        # Select the author
                        author_choices_prompt = inquirer.List(
                            'author_prop',
                            'Select the property to modify',
                            choices = list(keydictionary.keys())
                        )
                        answer = None
                        try:
                            answer = keydictionary[inquirer.prompt([author_choices_prompt])['author_prop']]
                        except KeyError:
                            continue
                        print(answer)
                        if answer == None:
                            continue
                        elif answer == Author.NAME:
                            result, answer = change_text(answer, author_dict[answer])
                            if result:
                                author.name = answer
                        elif answer == Author.TYPE:
                            author_type_choices = inquirer.List(
                                'author_type',
                                'Select the type of author to be changed to',
                                choices= [atype.value for atype in AuthorType]
                            )
                            try:
                                new_author_type = AuthorType(inquirer.prompt([author_type_choices])['author_type'])
                            except KeyError:
                                continue
                            author.type = new_author_type
                        project_config.author = author
                    is_modified = True
# manage the project
# Project.cli_manage()



import colorama
import copy
from datetime import datetime
import errno
import inquirer
import os
import pathlib
from typing import Any, Dict, List
import yaml
from yaml.error import YAMLError

from lexicon.commonutils import CLI_VARS_PRETTYPRINT, change_text, check_is_modified
from lexicon.makegenerators import (
    append_lines,
    gen_makefile_module_cmd,
    MAKEFILE_PHONY,
    PHONY_BUILD_ALL,
    PHONY_CLEAN_ALL,
)

from lexicon.module import Module


class Subproject:
    # --- INSTANCE VARIABLES
    name: str
    parent: str
    author: str
    modules: List[Module]

    # --- CLASS VARIABLES
    # - DICT (HASH-MAP) VARIABLES
    NAME = "Name"
    PARENT = "Parent"
    AUTHOR = "Author"
    MODULES = "Modules"
    # - FILENAME
    SUBPROJECT_FILE = "lexicon_sub.yaml"
    LEXICON_SUB_FILE_BACKUP = "." + SUBPROJECT_FILE + ".bkp"

    def __init__(self, args: Dict[str, Any]) -> None:
        # TODO: Argument Validation
        self.name = args[Subproject.NAME].strip()
        self.parent = args[Subproject.PARENT].strip()
        self.author = args[Subproject.AUTHOR]
        self.modules = []
        modules = args[Subproject.MODULES]
        # print(modules)
        for module_name in modules:
            self.modules.append(Module(module_name, modules[module_name]))

    def to_dict(self):
        modules = {}
        for module in self.modules:
            # small modification done here so that to_yaml() can
            # work properly, though that is not really needed for us
            modules[module.name] = module.to_dict()[module.name]
        return {
            Subproject.NAME: self.name,
            Subproject.PARENT: self.parent,
            Subproject.AUTHOR: self.author,
            Subproject.MODULES: modules
        }

    def to_yaml(self):
        return yaml.safe_dump(self.to_dict())

    def save_config(self, disable_replace = False):
        backup_path = os.path.join(os.getcwd(), Subproject.LEXICON_SUB_FILE_BACKUP)
        subproject_path = os.path.join(os.getcwd(), Subproject.SUBPROJECT_FILE)
        # Backup not implemented, unreachable code for now
        if not disable_replace:
            os.replace(subproject_path, backup_path)
        with open(subproject_path, "w") as new_config_file:
            yaml.safe_dump(self.to_dict(), new_config_file)

    def restore_config():
        backup_path = os.path.join(os.getcwd(), Subproject.LEXICON_SUB_FILE_BACKUP)
        lexicon_path = os.path.join(os.getcwd(), Subproject.SUBPROJECT_FILE)
        if os.path.exists(backup_path):
            os.replace(backup_path, lexicon_path)

    def get_files_not_in_directory(self) -> List[str]:
        files_not_in_dir = []
        for module in self.modules:
            files_not_in_dir += module.get_files_not_in_directory()
        return files_not_in_dir

    def create_files_not_in_directory(self):
        for file in self.get_files_not_in_directory():
            os.mknod(file)

    def create_build_list(self):
        build_modules = []
        for module in self.modules:
            build_modules.append(module.build_command)
        return build_modules

    def create_module_build_list(self):
        build_cmds = map(gen_makefile_module_cmd, self.create_build_list())
        return append_lines(PHONY_BUILD_ALL, *build_cmds)

    def create_clean_list(self):
        clean_modules = []
        for module in self.modules:
            clean_modules.append(module.clean_command)
        return clean_modules

    def create_module_clean_list(self):
        clean_cmds = map(gen_makefile_module_cmd, self.create_clean_list())
        return append_lines(PHONY_CLEAN_ALL, *clean_cmds)

    def create_makefile_overall_lists(self):
        return append_lines(
            MAKEFILE_PHONY,
            self.create_module_build_list(),
            self.create_module_clean_list()
        )

    def create_module_recipe_list(self):
        module_recipes = []
        for module in self.modules:
            module_recipes.append(f"# --- MODULE {module.name} ---")
            module_recipes.append(module.gen_complete_make_cmd())
        return append_lines(*module_recipes)

    def generate_makefile(self):
        return append_lines(
            "# MAKEFILE GENERATED BY LEXICON",
            f"# {Subproject.NAME}: {self.name}",
            f"# {Subproject.PARENT}: {self.parent}",
            f"# {Subproject.AUTHOR}: {self.author[:160]}",
            self.create_makefile_overall_lists(),
            self.create_module_recipe_list()
        ).strip()

    def write_makefile(self, path: str):
        try:
            with open(path, "w") as makefile_obj:
                makefile_obj.write(self.generate_makefile())
            print(colorama.Fore.GREEN + "Makefile Generated." + colorama.Fore.RESET)
        except FileExistsError as fee:
            print(colorama.Fore.RED + "Error in writing Makefile" +  colorama.Fore.RESET)
            raise fee
        except OSError as ose:
            print(colorama.Fore.RED + "Error in writing Makefile" +  colorama.Fore.RESET)
            raise ose


    # subroutine to generate a subproject
    @staticmethod
    def generate_subproject(
        subproject_name: str, 
        parent_name: str,
        author: str
    ):
        os.chdir(os.path.join(os.getcwd(), subproject_name))
        if os.path.exists(os.path.join(os.getcwd(), "lexicon_sub.yaml")):
            if inquirer.confirm("A project file exists here. Create a new one?"):
                _ = Subproject({
                    Subproject.NAME: subproject_name,
                    Subproject.PARENT: parent_name,
                    Subproject.AUTHOR: author,
                    Subproject.MODULES: []
                }).save_config(disable_replace=True)
        else:
            _ = Subproject({
                Subproject.NAME: subproject_name,
                Subproject.PARENT: parent_name,
                Subproject.AUTHOR: author,
                Subproject.MODULES: []
            }).save_config(disable_replace=True)
        Subproject.cli_manage()
        os.chdir(pathlib.Path(*pathlib.Path(os.getcwd()).parts[:-1]))


    @staticmethod
    def transfer_control(subproject_name: str, is_check_needed: bool=True):
        # check if we are in a project directory
        if os.path.exists(os.path.join(os.getcwd(), "lexicon.yaml")) or not is_check_needed:
            # we are in a project directory
            try:
                ans = None
                # change to the subproject directory
                os.chdir(os.path.join(os.getcwd(), subproject_name))
                # if the subproject file exists
                if os.path.exists(os.path.join(os.getcwd(), "lexicon_sub.yaml")):
                    # manage the project
                    ans = Subproject.cli_manage()
                    # get back to the project directory
                else:
                    # error that the subproject file can't be found
                    print(colorama.Fore.RED + "Project configuration file doesn't exist." + colorama.Fore.RESET)
                # go back to the parent directory
                # TODO: USE FIND AND REIMPLEMENT ALL PLACES OF PATH MANIPULATION USING PATHLIB
                os.chdir(pathlib.Path(*pathlib.Path(os.getcwd()).parts[:-1]))
                print(os.getcwd())
                return ans
            except OSError as ose:
                raise ose

    # -- VARIABLES FOR THE STATIC METHOD
    CHOICE = "choice"
    
    # CLI Management
    @staticmethod
    def cli_manage():
        # root prompt for 'Subproject'
        sub_filepath = os.path.join(os.getcwd(), Subproject.SUBPROJECT_FILE)
        if not os.path.exists(sub_filepath):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), sub_filepath)
        try:
            subproject_config = yaml.safe_load(open(sub_filepath, "r"))
            subproject_config: 'Subproject' = Subproject(subproject_config)
            reset_subconfig = copy.deepcopy(subproject_config)
            is_modified = False
            # CLI Loop
            while True:
                # ROOT PROMPT
                root_prompt = inquirer.List(
                    Subproject.CHOICE,
                    message=f"Manage {colorama.Fore.YELLOW}Subproject {colorama.Fore.BLUE}{subproject_config.name} {check_is_modified(is_modified,f'{colorama.Fore.GREEN}<modified>')}{colorama.Fore.RESET}",
                    choices=["Modify Properties", "Manage Modules", "Verify", "Reset", "Generate Makefile", "Exit"]
                )
                choice = inquirer.prompt([root_prompt])[Module.CHOICE]
                # ALL CHOICES ARE SELECTED HERE
                # If we're exiting it, make sure we do it properly.
                if choice == "Exit":
                    if is_modified:
                        modify_write = inquirer.confirm("Subproject has been changed. Do you want to save changes?")
                        if modify_write:
                            subproject_config.save_config()
                    return is_modified
                elif choice == "Generate Makefile":
                    # THESE ARE THE SUBROUTINES NECESSARY TO GENERATE MAKEFILES PROPERLY
                    makefile_path = os.path.join(os.path.dirname(sub_filepath), "Makefile")
                    # check if makefile exists
                    if os.path.exists(makefile_path):
                        # if so, create backup if needed
                        backup_confirmation = inquirer.confirm("Do you want to create a backup of your Makefile?")
                        if backup_confirmation:
                            backup_makefile_path = os.path.join(
                                os.path.dirname(sub_filepath),
                                "backups",
                                "Makefile_" + datetime.isoformat(datetime.now()) + ".bkp"
                            )
                            if not os.path.exists(os.path.dirname(backup_makefile_path)):
                                os.makedirs(os.path.dirname(backup_makefile_path))
                            os.replace(makefile_path, backup_makefile_path)
                        else:
                            os.remove(makefile_path)
                    subproject_config.write_makefile(makefile_path)
                elif choice == "Manage Modules":
                    while True:
                        module_map = {}
                        for module in subproject_config.modules:
                            module_map[module.name] = copy.deepcopy(module)
                        module_list = list(module_map.keys()) + [
                            "Create a module", 
                            "Exit"
                          ]
                        module_prompt = inquirer.List(
                            'module_choice',
                            'Select a module or an action',
                            choices = module_list
                        )
                        module_choice = inquirer.prompt([module_prompt])['module_choice']
                        if module_choice == "Exit":
                            break
                        elif module_choice == "Create a module":
                            module = Module.generate_module(subproject_config.name, subproject_config.parent)
                            subproject_config.modules.append(module)
                            is_modified = True
                        else:
                            selected_module: Module = module_map[module_choice]
                            selected_module_index = None
                            i = 0
                            for module in subproject_config.modules:
                                if selected_module.name == module.name:
                                    selected_module_index = i
                                i += 1
                            is_module_modified, result = Module.cli_manage(selected_module, subproject_config.name, subproject_config.parent)
                            if is_module_modified:
                                subproject_config.modules[selected_module_index] = result
                                is_modified = True
                                print(f"Modifications to {colorama.Fore.BLUE}{selected_module.name}{colorama.Fore.RESET} applied to {colorama.Fore.YELLOW}{subproject_config.name}{colorama.Fore.RESET}")
                elif choice == "Verify":
                    is_verified = None
                    files_not_in_dirs = []
                    for module in subproject_config.modules:
                        if module.verify():
                            # print(f"{colorama.Fore.GREEN}The module {module.name} has no errors.{colorama.Fore.RESET}")
                            if is_verified != False:
                                is_verified = True
                        else:
                            # print(f"{colorama.Fore.RED}The module {module.name} has errors!{colorama.Fore.RESET}")
                            files_not_in_dirs += module.get_files_not_in_directory()
                            is_verified = False
                    if is_verified == True:
                        print(f"{colorama.Fore.GREEN}The subproject {subproject_config.name} has no errors.{colorama.Fore.RESET}")
                    else:
                        print(f"{colorama.Fore.RED}The subproject {subproject_config.name} has errors.{colorama.Fore.RESET}")
                elif choice == "Reset":
                    reset_confirm = inquirer.confirm("Do you want to reset the subproject configuration?")
                    if reset_confirm:
                        subproject_config = reset_subconfig
                        is_modified = False
                elif choice == "Modify Properties":
                    subproject_dict = subproject_config.to_dict()
                    keydictionary = {}
                    for var in [
                        Subproject.NAME,
                        Subproject.PARENT,
                        Subproject.AUTHOR
                    ]:
                        keydictionary[CLI_VARS_PRETTYPRINT.format(var, subproject_dict[var])] = var
                    choice_list = inquirer.List(
                        'subproject_key',
                        message='Select the property to modify',
                        choices = list(keydictionary.keys())
                    )
                    choice_property = keydictionary[inquirer.prompt([choice_list])['subproject_key']]
                    result, answer = change_text(choice_property, subproject_dict[choice_property])
                    if result:
                        if choice_property == Subproject.NAME:
                            subproject_config.name = answer
                        elif choice_property == Subproject.PARENT:
                            subproject_config.parent = answer
                        elif choice_property == Subproject.AUTHOR:
                            subproject_config.author = answer
                        is_modified = True
        except YAMLError as ye:
            raise ye
        except OSError as ose:
            raise ose



# subproject_config = yaml.safe_load(open(os.path.join(os.getcwd(), "lexicon_sub.yaml"), "r"))

# subpro = Subproject(subproject_config)
# print(subpro.generate_makefile())

# Subproject.cli_manage()


import os

from lexicon.project import Project
from lexicon.subproject import Subproject

PROJECT_PATH = os.path.join(os.getcwd(), Project.PROJECT_FILE)
SUBPROJECT_PATH = os.path.join(os.getcwd(), Subproject.SUBPROJECT_FILE)


def main():
    if os.path.exists(SUBPROJECT_PATH):
        Subproject.cli_manage()
    elif os.path.exists(PROJECT_PATH):
        Project.cli_manage()
    else:
        Project.cli_generate()




import setuptools

with open("docs/README.md", "r", encoding="utf-8") as readmefile:
    long_description = readmefile.read()

# Set up all things
setuptools.setup(
    name = "lexicon-gauthamkrishna9991",
    version = "0.1.0",
    author = "Goutham Krishna K V",
    author_email = "gauthamkrishna9991@live.com",
    description = "A Verilog Project Management Tool",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/gauthamkrishna9991/lexicon",
    project_urls = {
        "Issues": "https://github.com/gauthamkrishna9991/issues"
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    package_dir = {
        "": "src"
    },
    install_requires = [
        "blessed==1.17.6",
        "colorama==0.4.4",
        "inquirer==2.7.0",
        "python-editor==1.0.4",
        "PyYAML==5.4.1",
        "readchar==2.0.1",
        "six==1.15.0",
        "wcwidth==0.2.5",
    ],
    entry_points = {
        "console_scripts": [
            "lexicon=lexicon.main:main"
        ]
    },
    packages = setuptools.find_packages(where="src"),
    python_requires = ">= 3.6"
)


import copy
import colorama
import glob
import inquirer
import os
from typing import Any, Dict, List, Tuple
import yaml


from lexicon.commonutils import (
    change_text,
    check_is_modified,
    CLI_VARS_PRETTYPRINT
)
from lexicon.filegenerators import generate_quickstart_template
from lexicon.makegenerators import (
    COMPILE_COMMAND,
    CLEAN_COMMAND,
    VIEWER_COMMAND,
    append_lines,
    gen_executable,
    gen_inputvar_const,
    gen_input_output,
    gen_report_cmd
)


def remove_files_in_directory(files: List[str],extension: str = "*.v"):
    dirfiles = glob.glob(extension, recursive=False)
    for file in dirfiles:
        if file in files:
            files.remove(file)
    return files

#---------------------------------------------------------------------------------------
#TODO: CHANGE ALL THE ASSERTIONS TO CONDITIONAL ValueError Exceptions.... DO NOT FORGET|
#---------------------------------------------------------------------------------------
class Module:
    # --- CLASS MEMBERS (INSTANCE VARIABLES) ---
    # module name
    name: str
    # module file
    files: List[str]
    files_var: str
    # module executable
    exec_file: str
    exec_filevar: str
    # module wave files (with '.vcd' extension)
    wave_file: str
    wave_filevar: str
    # build command
    build_command: str
    # clean command
    clean_command: str
    # timeout for the executable
    timeout: int

    # --- CLASS VARIABLES ---
    # - CONSTANTS NEEDED FOR HASH-MAP RESOLUTION
    FILES = 'files'
    FILES_VAR = 'files_var'
    EXEC = 'exec'
    EXEC_VAR = 'exec_var'
    WAVE = 'wave'
    WAVE_VAR = 'wave_var'
    BUILD_COMMAND = 'build_command'
    CLEAN_COMMAND = 'clean_command'
    TIMEOUT = "timeout"
    # - ERROR MESSAGES
    STR_ERR = "{} of {} must be a string"
    FILE_ERR = "{} of {} must be a list of files (as strings)"
    FILENAME_ERR = "variable '{}' ({}) in {} MUST NOT have spaces"
    # DEFAULT VARIABLES
    # TODO: Move this to a user/project configuration file when needed.
    # default timeout: 10 seconds
    DEFAULT_TIMEOUT = 10

    # --- METHODS ---
    # constructor for a specific module
    def __init__(self, name: str, args: Dict[str,Any]) -> None:
        # ALL ASSERTIONS IN THIS FILE ARE MEANT TO ENSURE PROPER MAKEFILE GENERATION
        # name
        name = name.strip()
        if " " in name:
            raise ValueError(f"Change the name of module '{name}' so it doesn't have any spaces in it.")
        if name != name.lower():
            raise ValueError(f"The module '{name}' must not contain capital letters. Change it.")
        # ASSERTIONS ADDED FOR DEBUG
        # assert(" " not in name), f"change the name of module '{name}' so it doesn't have any spaces in it."
        # assert(name == name.lower()), f"the name of module '{name}' contains capital letters. Change it."
        self.name = name
        # files
        if args == None:
            args = {}
        if Module.FILES in args:
            if args[Module.FILES] is None:
                self.files = []
            if type(args[Module.FILES]) != list:
                print(args[Module.FILES])
                raise ValueError(Module.FILE_ERR.format(Module.FILES, self.name))
            # ASSERTIONS ADDED AT THE TIME OF DEBUG
            # assert (type(args[Module.FILES]) == list or type(args[Module.FILES] == None)), Module.FILE_ERR.format(Module.FILES, self.name)
            for file in args[Module.FILES]:
                if type(file) != str:
                    raise ValueError(f"invalid filename '{file}' for {self.name}")
                # assert (type(file) == str), f"invalid filename '{file}' for {self.name}"
            self.files = args[Module.FILES]
        else:
            testbench_file: str = f"{self.name}_tb.v"
            module_file: str = f"{self.name}_mod.v"
            tempfiles = [testbench_file, module_file]
            print("Keys not given.")
            print("Defaulting to modules " + str(tempfiles))
            self.files = tempfiles
        # files_var
        if Module.FILES_VAR in args:
            if type(args[Module.FILES_VAR]) != str:
                raise ValueError(Module.STR_ERR.format(Module.files_var, self.files_var))
            if " " in args[Module.FILES_VAR]:
                raise ValueError(Module.FILENAME_ERR.format(args[Module.FILES_VAR], Module.files_var))
            assert (" " not in args[Module.FILES_VAR]), Module.FILENAME_ERR.format(args[Module.FILES_VAR], Module.FILES_VAR, self.name)
            self.files_var = args[Module.FILES_VAR]
        else:
            self.files_var = self.name.upper() + "_FILES"
        # exec
        if Module.EXEC in args:
            assert (type(args[Module.EXEC]) == str), Module.STR_ERR.format(Module.EXEC, self.name)
            self.exec_file = args[Module.EXEC]
        else:
            self.exec_file = self.name + ".vvp"
        # exec_var
        if Module.EXEC_VAR in args:
            assert (type(args[Module.EXEC_VAR]) == str), Module.STR_ERR.format(Module.EXEC_VAR, self.name)
            self.exec_filevar = args[Module.EXEC_VAR]
        else:
            self.exec_filevar = self.name.upper() + "_EXEC"
        # wave
        if Module.WAVE in args:
            assert (type(args[Module.WAVE]) == str), Module.STR_ERR.format(Module.WAVE, self.name)
            self.wave_file = args[Module.WAVE]
        else:
            self.wave_file = self.name + ".vcd"
        # wave_var
        if Module.WAVE_VAR in args:
            assert (type(args[Module.WAVE_VAR]) == str), Module.STR_ERR.format(Module.WAVE_VAR, self.name)
            self.wave_filevar =  args[Module.WAVE_VAR]
        else:
            self.wave_filevar = self.name.upper() + "_WAVE"
        # build command
        if Module.BUILD_COMMAND in args:
            assert (type(args[Module.BUILD_COMMAND]) == str), Module.STR_ERR.format(Module.BUILD_COMMAND, self.name)
            self.build_command = args[Module.BUILD_COMMAND]
        else:
            # use the module name itself for build
            self.build_command = self.name.lower()
        # clean command
        if Module.CLEAN_COMMAND in args:
            assert(type(args[Module.CLEAN_COMMAND]) == str), Module.STR_ERR.format(Module.BUILD_COMMAND, self.name)
            self.clean_command = args[Module.CLEAN_COMMAND]
        else:
            # name + "_clean" for the clean job in makefile
            self.clean_command = self.name + "_clean"
        # timeout
        if Module.TIMEOUT in args:
            assert(type(args[Module.TIMEOUT]) == int), f"timeout of {self.name} must be an integer"
            try:
                self.timeout = int(args[Module.TIMEOUT])
            except Exception as e:
                raise e
        else:
            self.timeout = Module.DEFAULT_TIMEOUT

    # generate a dictionary (hash-map) object for representation
    def to_dict(self):
        return {
            self.name: {
                Module.FILES: self.files,
                Module.FILES_VAR: self.files_var,
                Module.EXEC: self.exec_file,
                Module.EXEC_VAR: self.exec_filevar,
                Module.WAVE: self.wave_file,
                Module.WAVE_VAR: self.wave_filevar,
                Module.BUILD_COMMAND: self.build_command,
                Module.CLEAN_COMMAND: self.clean_command,
                Module.TIMEOUT: self.timeout
            }
        }

    # generate yaml file back
    def to_yaml(self):
        return yaml.safe_dump(self.to_dict())
    
    def verify(self):
        return remove_files_in_directory(self.files) == []

    def create_files_not_in_directory(self):
        for file in self.get_files_not_in_directory():
            os.mknod(file)
    
    def get_files_not_in_directory(self):
        return remove_files_in_directory(self.files, "*.v")
    
    # --- GENERATION METHODS

    # A FILE VARIABLE GENERATION
    #     - exec file generation
    def gen_exec_file(self):
        return f"{self.exec_filevar} = {self.exec_file}"
    
    #     - wave file generation
    def gen_wave_file(self):
        return f"{self.wave_filevar} = {self.wave_file}"

    #     - files variable generation
    def gen_files_var(self):
        comment = f"# {self.name}: VARS"
        files_var = f"{self.files_var} ="
        for file in self.files:
            files_var += " " + file
        return comment + "\n" + files_var
    
    # a OVERALL FILE DEFENITION GENERATION
    def gen_file_cmd(self):
        return append_lines(self.gen_files_var(), self.gen_exec_file(), self.gen_wave_file())

    # B BUILD PROCESS GENERATION
    #   - compile process generation
    def gen_compile_cmd(self):
        return append_lines(
            f"# {self.name}: BUILD -> COMPILE",
            gen_input_output(self.files_var, self.exec_filevar),
            gen_report_cmd(self.name, "Compiling Source Files..."),
            COMPILE_COMMAND
        )
    
    #   - execute command generation
    def gen_execute_cmd(self):
        return append_lines(
            f"# {self.name}: BUILD -> EXECUTE",
            gen_input_output(self.exec_filevar, self.wave_filevar),
            gen_report_cmd(self.name, f"Running Simulation with timeout \\e[1;93m{self.timeout}s"),
            gen_executable(self.timeout)
        )

    #   - view command generation
    def gen_viewer_cmd(self):
        return append_lines(
            f"# {self.name}: BUILD -> VIEW",
            gen_inputvar_const(self.wave_filevar, self.build_command),
            gen_report_cmd(self.name, "Running Simulation"),
            VIEWER_COMMAND
        )

    # b - whole build + execute process
    def gen_build_cmd(self):
        return append_lines(
            f"# {self.name}: BUILD",
            self.gen_viewer_cmd(),
            self.gen_execute_cmd(),
            self.gen_compile_cmd()
        )

    # B CLEAN PROCESS GENERATION
    def gen_clean_cmd(self):
        return append_lines(
            f"# {self.name}: CLEAN",
            f"{self.clean_command}:",
            gen_report_cmd(self.name, "Cleaning files of the module"),
            CLEAN_COMMAND.format(self.exec_filevar, self.wave_filevar)
        )

    def gen_complete_make_cmd(self):
        return append_lines(
            self.gen_file_cmd(),
            self.gen_build_cmd(),
            self.gen_clean_cmd()
        )

    # --- VARIABLES FOR THE STATIC METHOD
    CHOICE = "choice"

    # --- CLI METHODS ---
    def generate_module(subproject_name: str, parent_name: str):
        module_prompt = inquirer.Text(
            'modname',
            'Enter the name of the new module'
        )
        modname = inquirer.prompt([module_prompt])['modname']
        if modname == None:
            raise TypeError(modname)
        else:
            new_module = Module(modname, None)
            new_module.create_files_not_in_directory()
            new_module.cli_manage(new_module, subproject_name, parent_name)
            return new_module


    #   - CLI specific variables
    CLI_VARS_PRETTYPRINT: str = colorama.Fore.BLUE + "{}: " + colorama.Fore.YELLOW +"{}" + colorama.Fore.RESET
    def get_vars_prettyprint(choice: str):
        values = choice.split(":")
        return (values[0].strip(), values[1].strip())

    @staticmethod
    def cli_manage(
        module: 'Module',
        subproject_name: str,
        project_name: str
    ) -> Tuple[bool, 'Module']:
        reset_module = copy.deepcopy(module)
        is_modified = False   
        if module.files == []:
            generate_res = inquirer.confirm("Module files are empty. Do you want to generate default module files?")
            if generate_res:
                module.files = generate_quickstart_template(module, subproject_name, project_name)
                is_modified = True
        if module is None:
            # TODO: add a conditional module generator here
            raise TypeError(module)
        if subproject_name is None:
            raise UserWarning("Subproject name is not given")
        if project_name is None:
            raise UserWarning("Project name is not given")
        while (True):
            root_prompt = inquirer.List(
                Module.CHOICE,
                message=f"Manage Module {colorama.Fore.BLUE}{module.name} {check_is_modified(is_modified, f'{colorama.Fore.GREEN}<modified>')}{colorama.Fore.RESET}",
                choices=["Modify Properties", "Add or Remove Files", "Reset", "Verify", "Exit"]
            )
            # print(module.to_yaml())
            # ASK THE QUESTION
            choice = inquirer.prompt([root_prompt])[Module.CHOICE]
            # IF THE ANSWER IS EXIT
            if choice.strip() == "Exit":
                return (is_modified, module)
            # IF THE ANSWER IS TO ADD OR REMOVE FILES
            elif choice == "Add or Remove Files":
                # print("FILES IN DIR: " + str(glob.glob("*.v")))
                # print("FILES IN MODULE: " + str(module.files))
                file_select_inquiry = [
                    inquirer.Checkbox(
                        "files",
                        message="Select the files needed for selection (use space bar to select or deselect an object)",
                        choices=glob.glob("*.v", recursive=False),
                        default=module.files
                    )
                ]
                response = inquirer.prompt(file_select_inquiry)["files"]
                # print("FILES AFTER RESPONSE: " + str(response))
                # print("FILES IN MODULE AFTER RESPONSE: " + str(module.files))
                if response != module.files:
                    module.files = response
                    is_modified = True
            # IF WE NEED TO RESET THE VARIABLES
            elif choice == "Reset":
                reset_prompt = inquirer.confirm("Do you want to discard all changes?")
                if reset_prompt:
                    module = copy.deepcopy(reset_module)
                    is_modified = False
            # IF WE NEED TO VERIFY THE RESOURCES
            elif choice == "Verify":
                if module.get_files_not_in_directory() == []:
                    print(colorama.Fore.GREEN + "Module is verified to not have any problems." + colorama.Fore.RESET)
                else:
                    print(colorama.Fore.RED + "The module has some files which do not yet exist." + colorama.Fore.RESET)
                    print(f"Files not in directory: {module.get_files_not_in_directory()}")
                    do_reset = inquirer.confirm("Do you want to generate these files?")
                    if do_reset:
                        module.create_files_not_in_directory()
                        print(f"All files for '{colorama.Fore.BLUE}{module.name}{colorama.Fore.RESET}' created.")
            elif choice == "Modify Properties":
                module_dict = module.to_dict()[module.name]
                keydictionary = {}
                for var in [
                    Module.FILES_VAR,
                    Module.EXEC,
                    Module.EXEC_VAR,
                    Module.WAVE,
                    Module.WAVE_VAR,
                    Module.BUILD_COMMAND,
                    Module.CLEAN_COMMAND,
                    Module.TIMEOUT
                ]:
                    keydictionary[CLI_VARS_PRETTYPRINT.format(var,  module_dict[var])] = var
                choice_list = inquirer.List(
                    'module_key',
                    message='Select the property to change',
                    choices = list(keydictionary.keys())
                )
                choice_property = keydictionary[inquirer.prompt([choice_list])['module_key']]
                result, answer = change_text(choice_property, module_dict[choice_property])
                if result:
                    if choice_property == Module.FILES_VAR:
                        module.files_var = answer
                    elif choice_property == Module.EXEC:
                        module.exec_file = answer
                    elif choice_property == Module.EXEC_VAR:
                        module.exec_filevar = answer
                    elif choice_property == Module.WAVE:
                        module.wave_file = answer
                    elif choice_property == Module.WAVE_VAR:
                        module.wave_filevar = answer
                    elif choice_property == Module.BUILD_COMMAND:
                        module.build_command = answer
                    elif choice_property == Module.CLEAN_COMMAND:
                        module.clean_command = answer
                    elif choice_property == Module.TIMEOUT:
                        module.timeout = int(answer)
                    is_modified = True


from re import match
from typing import Tuple
import inquirer
import colorama


def change_text(property: str, value: str) -> Tuple[bool, str]:
    text_modify_question = inquirer.Text(
        name="property",
        message=f"Modify {colorama.Fore.BLUE}{property}{colorama.Fore.RESET}",
        default=value
    )
    answer = inquirer.prompt([text_modify_question])["property"].strip()
    if inquirer.confirm(f"Are you sure you want to change {colorama.Fore.BLUE}{property}{colorama.Fore.RESET} to {colorama.Fore.YELLOW}{answer}{colorama.Fore.RESET}?" ):
        return (True, answer)
    return (False, None)

def change_int(property: str, value: int) -> Tuple[bool, int]:
    int_modify_question = inquirer.Text(
        name="property",
        message=f"Modify {colorama.Fore.BLUE}{property}{colorama.Fore.RESET}",
        default=value,
        validate= lambda _, x: match('\d[\d]+', x)
    )
    try:
        answer = int(inquirer.prompt(int_modify_question)["property"].strip())
        if inquirer.confirm(f"Are you sure you want to change {colorama.Fore.BLUE}{property}{colorama.Fore.RESET} to {colorama.Fore.YELLOW}{answer}{colorama.Fore.RESET}"):
            return (True, answer)
        return (False, None)
    except ValueError as ve:
        raise ve

def check_is_modified(val, out):
    if val:
        return out
    return ""

# --- CLI SPECIFIC VARIABLES
CLI_VARS_PRETTYPRINT: str = colorama.Fore.BLUE + "{}: " + colorama.Fore.YELLOW +"{}" + colorama.Fore.RESET
# def get_vars_prettyprint(choice: str):
#     values = choice.split(":")
#     return (values[0].strip(), values[1].strip())